@@ -1,11 +1,13 @@
+import asyncio
 import inspect
 import json
 import logging
 import random
 import re
 import sys
+import time
 from collections import defaultdict
-from contextlib import redirect_stdout
+from contextlib import contextmanager, redirect_stdout
 from datetime import datetime, timedelta
 from itertools import product
 from os import getenv
@@ -22,7 +24,7 @@ import numpy as np
 import pandas as pd
 import pytz
 from dateutil.relativedelta import relativedelta
-from humanize import naturaltime
+from humanize import naturaldelta, naturaltime
 from sqlalchemy import Date, and_, cast, desc
 from unidecode import unidecode
 
@@ -44,10 +46,12 @@ DEFAULT_CONFIG_CHANNELS = RUNTIME_ROOT / "channels.txt"
 TMP_DIR = RUNTIME_ROOT / "tmp"
 GRAPHCMD_FILE = TMP_DIR / "graphcmd.png"
 LASTWEEKCMD_FILE = TMP_DIR / "lastweek.png"
+EXPORT_FILE = TMP_DIR / "export.json"
 MIGRATIONS_FILE = DB_DIR / "migrations.txt"
 MIGRATIONS_DIR = SCRIPTS_DIR / "migrations"
 
 # Application Settings
+METRICS_INTERVAL = 15  # seconds
 EMBED_LIMIT = 5  # more embeds in a row than this causes issues
 BASE_PROPHET_URL = "https://turnipprophet.io/?prices="
 DAYS = {  # Based on values from datetime.isoweekday()
@@ -62,10 +66,17 @@ DAYS = {  # Based on values from datetime.isoweekday()
 IDAYS = dict(map(reversed, DAYS.items()))  # Reverse lookup from DAYS dict
 
 
-class PrefValidate:
-    """Provides utilites for validating user input of preferences"""
+class Direct:
+    """Use this to indicate that a command should send a DM."""
 
-    PREFRENCES = [
+    def __init__(self, response):
+        self.response = response
+
+
+class Preferences:
+    """Provides utilites for validating user input of preferences."""
+
+    KEYS = [
         "hemisphere",
         "timezone",
         "island",
@@ -80,7 +91,7 @@ class PrefValidate:
     @classmethod
     def key(cls, value):
         value = value.lower()
-        if value in cls.PREFRENCES:
+        if value in cls.KEYS:
             return value
         return {
             "code": "friendcode",
@@ -131,9 +142,15 @@ class PrefValidate:
     def island(cls, value):
         return value
 
+    @classmethod
+    def validate(cls, key, value):
+        return getattr(cls, key)(value)
+
 
 def h(dt):
     """Convertes a datetime to something readable by a human."""
+    if isinstance(dt, timedelta):
+        return naturaldelta(dt)
     if hasattr(dt, "tz_convert"):  # pandas-datetime-like objects
         dt = dt.to_pydatetime()
     naive_dt = dt.replace(tzinfo=None)
@@ -214,8 +231,13 @@ def discord_user_from_name(channel, name):
     if name is None:
         return None
     lname = name.lower()
-    members = channel.members
-    return next(filter(lambda member: lname in str(member).lower(), members), None)
+    if hasattr(channel, "members"):  # channels
+        members = channel.members
+        return next(filter(lambda member: lname in str(member).lower(), members), None)
+    else:  # direct messages
+        recipient = channel.recipient
+        rvalue = recipient if name in str(recipient).lower() else None
+        return rvalue
 
 
 def discord_user_from_id(channel, user_id):
@@ -223,8 +245,12 @@ def discord_user_from_id(channel, user_id):
     if user_id is None:
         return None
     iid = int(user_id)
-    members = channel.members
-    return next(filter(lambda member: iid == member.id, members), None)
+    if hasattr(channel, "members"):  # channels
+        members = channel.members
+        return next(filter(lambda member: iid == member.id, members), None)
+    else:  # direct messages
+        recipient = channel.recipient
+        return recipient if recipient.id == user_id else None
 
 
 def discord_user_name(channel, name_or_id):
@@ -243,7 +269,8 @@ def discord_user_id(channel, name):
     """Returns the discord user id name from the given channel and name."""
     if not name:
         return None
-    return getattr(discord_user_from_name(channel, name), "id", None)
+    user = discord_user_from_name(channel, name)
+    return user.id if user else None
 
 
 def is_turbot_admin(channel, user_or_member):
@@ -271,10 +298,17 @@ def command(f):
 class Turbot(discord.Client):
     """Discord turnip bot"""
 
-    def __init__(self, token="", channels=[], db_url=DEFAULT_DB_URL, log_level=None):
-        if log_level:  # pragma: no cover
-            logging.basicConfig(level=log_level)
-        super().__init__()
+    def __init__(
+        self,
+        token="",
+        channels=[],
+        db_url=DEFAULT_DB_URL,
+        log_level=logging.ERROR,
+        metrics=False,
+    ):
+        logging.basicConfig(level=log_level)
+        loop = asyncio.get_event_loop()
+        super().__init__(loop=loop)
         self.token = token
         self.channels = channels
         self.last_backup_filename = None
@@ -299,9 +333,44 @@ class Turbot(discord.Client):
         # load static application data
         self.assets = Assets()
 
+        # begin metrics logging
+        self._metrics = None
+        if metrics:  # pragma: no cover
+            loop.create_task(self._begin_metrics())
+
     def run(self):  # pragma: no cover
         super().run(self.token)
 
+    async def _begin_metrics(self):  # pragma: no cover
+        if not self._metrics:
+            self._metrics = logging.getLogger("turbot-metrics")
+            self._metrics.setLevel(logging.INFO)
+        while True:
+            data = {"is_ready": self.is_ready(), "guilds#len": len(self.guilds)}
+            metrics = [
+                m
+                for m in inspect.getmembers(self, lambda m: not inspect.isroutine(m))
+                if m[0].endswith("_spent_time")
+            ]
+            data.update(metrics)
+            self._metrics.info(json.dumps(data))
+            await asyncio.sleep(METRICS_INTERVAL)
+
+    @contextmanager
+    def _capture_metrics(self, kind):
+        start_time = time.time()
+        yield
+        spent_time = time.time() - start_time
+        if not hasattr(self, f"{kind}_max_spent_time") or spent_time > getattr(
+            self, f"{kind}_max_spent_time"
+        ):
+            setattr(self, f"{kind}_max_spent_time", spent_time)
+        if not hasattr(self, f"{kind}_avg_spent_time"):
+            setattr(self, f"{kind}_avg_spent_time", spent_time)
+        else:
+            avg = getattr(self, f"{kind}_avg_spent_time")
+            setattr(self, f"{kind}_avg_spent_time", (avg + spent_time) / 2.0)
+
     def backup_prices(self, data):
         """Backs up the prices data to a datetime stamped file."""
         filename = datetime.now(pytz.utc).strftime("prices-%Y-%m-%d.csv")
@@ -429,10 +498,11 @@ class Turbot(discord.Client):
 
     def generate_graph(self, channel, target_user, graphname):  # pragma: no cover
         """Generates a nice looking graph of user data."""
-        plot = self.get_graph(channel, target_user, graphname)
-        success = plot is not None
-        plt.close("all")
-        return success
+        with self._capture_metrics("graph"):
+            plot = self.get_graph(channel, target_user, graphname)
+            success = plot is not None
+            plt.close("all")
+            return success
 
     def append_price(self, author, kind, price, at):
         """Adds a price to the prices data file for the given author and kind."""
@@ -578,6 +648,11 @@ class Turbot(discord.Client):
                     raise
                 finally:
                     self.session.close()
+            if isinstance(response, Direct):
+                send_direct = True
+                response = response.response
+            else:
+                send_direct = False
             if not isinstance(response, list):
                 response = [response]
             last_reply_index = len(response) - 1
@@ -593,14 +668,20 @@ class Turbot(discord.Client):
                             and n == last_reply_index
                             else None
                         )
-                        await message.channel.send(page, file=file)
+                        if send_direct:
+                            await message.author.send(page, file=file)
+                        else:
+                            await message.channel.send(page, file=file)
                 elif isinstance(reply, discord.embeds.Embed):
                     file = (
                         attachment
                         if attachment is not None and n == last_reply_index
                         else None
                     )
-                    await message.channel.send(embed=reply, file=file)
+                    if send_direct:
+                        await message.author.send(embed=reply, file=file)
+                    else:
+                        await message.channel.send(embed=reply, file=file)
                 else:
                     raise RuntimeError("non-string non-embed reply not supported")
 
@@ -610,8 +691,14 @@ class Turbot(discord.Client):
 
     async def on_message(self, message):
         """Behavior when the client gets a message from Discord."""
-        # only respond in text channels
-        if str(message.channel.type) != "text":
+        # don't respond to any bots
+        if message.author.bot:
+            return
+
+        private = str(message.channel.type) == "private"
+
+        # only respond in text channels and to direct messages
+        if not private and str(message.channel.type) != "text":
             return
 
         # don't respond to yourself
@@ -622,24 +709,26 @@ class Turbot(discord.Client):
         if not message.content.startswith("!"):
             return
 
-        # check for admin authorized channels on this server
-        guild_id = message.channel.guild.id
-        rows = self.data.conn.execute(
-            f"SELECT * FROM authorized_channels WHERE guild = {guild_id};"
-        )
-        authorized_channels = set(row["name"] for row in rows)
-        if authorized_channels and message.channel.name not in authorized_channels:
-            return
+        if not private:
+            # check for admin authorized channels on this server
+            guild_id = message.channel.guild.id
+            rows = self.data.conn.execute(
+                f"SELECT * FROM authorized_channels WHERE guild = {guild_id};"
+            )
+            authorized_channels = set(row["name"] for row in rows)
+            if authorized_channels and message.channel.name not in authorized_channels:
+                return
 
-        # fallback to checking against channels authorized at application startup time
-        if (
-            not authorized_channels
-            and self.channels
-            and message.channel.name not in self.channels
-        ):
-            return
+            # fallback to checking against channels authorized at application startup time
+            if (
+                not authorized_channels
+                and self.channels
+                and message.channel.name not in self.channels
+            ):
+                return
 
-        await self.process(message)
+        with self._capture_metrics("process"):
+            await self.process(message)
 
     async def on_ready(self):
         """Behavior when the client has successfully connected to Discord."""
@@ -701,7 +790,7 @@ class Turbot(discord.Client):
             usage += f"\n>  {use}"
             usage += "\n"
         usage += "\n_Turbot created by TheAstropath_"
-        return usage, None
+        return Direct(usage), None
 
     class _PriceTimeError(Exception):
         def __init__(self, key):
@@ -808,6 +897,9 @@ class Turbot(discord.Client):
         Only Turbot Admin members can run this command. Generates a final graph for use
         with !lastweek and resets all data for all users.
         """
+        if str(channel.type) == "private":
+            return s("no_dm"), None
+
         if not is_turbot_admin(channel, author):
             return s("not_admin"), None
 
@@ -866,13 +958,13 @@ class Turbot(discord.Client):
         yours = prices[prices.author == target_id]
         lines = [s("history_header", name=target_name)]
         for _, row in yours.iterrows():
-            time = self.to_usertime(target_id, row.timestamp)
+            usertime = self.to_usertime(target_id, row.timestamp)
             lines.append(
                 s(
                     f"history_{row.kind}",
                     price=row.price,
-                    timestamp=h(time),
-                    day_and_time=day_and_time(time),
+                    timestamp=h(usertime),
+                    day_and_time=day_and_time(usertime),
                 )
             )
         return "\n".join(lines), None
@@ -907,10 +999,17 @@ class Turbot(discord.Client):
         The default is to look for the best sell.
         @ [buy|sell]
         """
+        if str(channel.type) == "private":
+            return s("no_dm"), None
+
         kind = params[0].lower() if params else "sell"
         if kind not in ["sell", "buy"]:
             return s("best_invalid_param"), None
 
+        fudge_minutes = 5  # it takes a while to get over to an island
+        commandtime = datetime.now(pytz.utc) + timedelta(minutes=fudge_minutes)
+        commandtime = self.to_usertime(author.id, commandtime)
+
         prices = self.data.prices
         past = datetime.now(pytz.utc) - timedelta(hours=12)
         sells = prices[(prices.kind == kind) & (prices.timestamp > past)]
@@ -920,8 +1019,28 @@ class Turbot(discord.Client):
         for _, row in bests.iterrows():
             name = discord_user_from_id(channel, row.author)
             if name:
-                timestamp = h(self.to_usertime(row.author, row.timestamp))
-                lines.append(s("best", name=name, price=row.price, timestamp=timestamp))
+                usertime = self.to_usertime(row.author, row.timestamp)
+                closing = datetime(
+                    year=usertime.year,
+                    month=usertime.month,
+                    day=usertime.day,
+                    hour=22 if kind == "sell" else 12,
+                    tzinfo=usertime.tzinfo,
+                )
+                if closing <= commandtime:
+                    continue  # sell/buy is no longer available anymore
+                remaining = closing - commandtime
+                lines.append(
+                    s(
+                        "best",
+                        name=name,
+                        price=row.price,
+                        timestamp=h(usertime),
+                        remaining=h(remaining),
+                    )
+                )
+        if len(lines) == 1:
+            lines.append(s("best_none_found"))
         return "\n".join(lines), None
 
     @command
@@ -1017,6 +1136,9 @@ class Turbot(discord.Client):
         the in-game item name, and more than one can be provided if separated by commas.
         @ <comma, separated, list, of, collectables>
         """
+        if str(channel.type) == "private":
+            return s("no_dm"), None
+
         if not params:
             return s("search_no_params"), None
 
@@ -1024,7 +1146,7 @@ class Turbot(discord.Client):
         valid, invalid = self.assets.validate(items)
 
         def get_results(kind, valid_items):
-            store = getattr(self.data, kind)
+            store = self.data[kind]
             users = store.author.unique()
             results = defaultdict(list)
             for collected_item in valid_items:
@@ -1081,7 +1203,7 @@ class Turbot(discord.Client):
         lines = []
 
         def add_lines(kind, fullset):
-            store = getattr(self.data, kind)
+            store = self.data[kind]
             your_items = store[store.author == target_id]
             collected_items = set(your_items.name.unique())
             remaining_items = fullset - collected_items
@@ -1107,6 +1229,9 @@ class Turbot(discord.Client):
         give the name of the kind of collectable to return.
         @ <fossils|bugs|fish|art|songs>
         """
+        if str(channel.type) == "private":
+            return s("no_dm"), None
+
         if not params:
             return s("needed_no_param"), None
 
@@ -1115,7 +1240,7 @@ class Turbot(discord.Client):
             return s("needed_invalid_param"), None
 
         fullset = self.assets[kind].all
-        store = getattr(self.data, kind)
+        store = self.data[kind]
         authors = [member.id for member in channel.members if member.id != self.user.id]
         total = pd.DataFrame(list(product(authors, fullset)), columns=store.columns)
         merged = total.merge(store, indicator=True, how="outer")
@@ -1153,7 +1278,7 @@ class Turbot(discord.Client):
             return s("cant_find_user", name=target), None
 
         def get_collection(kind):
-            store = getattr(self.data, kind)
+            store = self.data[kind]
             your_items = store[store.author == target_id]
             return set(your_items.name.unique())
 
@@ -1212,16 +1337,16 @@ class Turbot(discord.Client):
         Set one of your user preferences. @ <preference> <value>
         """
         if not params:
-            return s("pref_no_params", prefs=", ".join(PrefValidate.PREFRENCES)), None
+            return s("pref_no_params", prefs=", ".join(Preferences.KEYS)), None
 
-        pref = PrefValidate.key(params[0])
+        pref = Preferences.key(params[0])
         if not pref:
-            return s("pref_invalid_pref", prefs=", ".join(PrefValidate.PREFRENCES)), None
+            return s("pref_invalid_pref", prefs=", ".join(Preferences.KEYS)), None
         if len(params) <= 1:
             return s("pref_no_value", pref=pref), None
 
         value = " ".join(params[1:])
-        validated_value = getattr(PrefValidate, pref)(value)
+        validated_value = Preferences.validate(pref, value)
         if isinstance(validated_value, list):
             return s("did_you_mean", possible=", ".join(validated_value)), None
         if not validated_value:
@@ -1256,7 +1381,7 @@ class Turbot(discord.Client):
 
         def add_valid_lines(kind, fullset):
             lines.append(s(f"count_{kind}_valid_header"))
-            store = getattr(self.data, kind)
+            store = self.data[kind]
             for user_name, user_id in sorted(valid):
                 yours = store[store.author == user_id]
                 collected = set(yours.name.unique())
@@ -1327,11 +1452,11 @@ class Turbot(discord.Client):
     def creatures_available_now(self, now, df):
         """Returns the names of creatures in df that are available right now."""
         for _, row in df.iterrows():
-            time = row["time"]
-            if time.lower() == "all day":
+            available_time = row["time"]
+            if available_time.lower() == "all day":
                 yield row["name"]
             else:
-                ranges = time.lower().split("&")
+                ranges = available_time.lower().split("&")
                 for r in ranges:
                     lhs, rhs = [t.strip().lower() for t in r.split("-")]
                     lhs_hour, lhs_ampm = lhs.split(" ")
@@ -1630,16 +1755,27 @@ class Turbot(discord.Client):
         Set the list of channels where Turbot is authorized to respond.
         @ <your, list, of, channels>
         """
+        if str(channel.type) == "private":
+            return s("no_dm"), None
         if not is_turbot_admin(channel, author):
             return s("not_admin"), None
         names = [name.strip() for name in " ".join(params).split(",")]
-        session = self.data.Session()
-        session.query(AuthorizedChannel).filter_by(guild=channel.guild.id).delete()
+        self.session.query(AuthorizedChannel).filter_by(guild=channel.guild.id).delete()
         for name in names:
-            session.add(AuthorizedChannel(guild=channel.guild.id, name=name))
-        session.commit()
+            self.session.add(AuthorizedChannel(guild=channel.guild.id, name=name))
         return s("authorize", channels=", ".join(names)), None
 
+    @command
+    def export(self, channel, author, params):
+        """
+        Exports all your data. It will be sent to you in a Direct Message.
+        """
+        self.ensure_user_exists(author)
+        user = self.session.query(User).get(author.id)
+        with open(EXPORT_FILE, "w") as f, redirect_stdout(f):
+            print(json.dumps(user.to_dict()))
+        return Direct(s("export")), discord.File(EXPORT_FILE)
+
 
 def get_token(token_file):  # pragma: no cover
     """Returns the discord token from the environment or your token config file."""
@@ -1709,6 +1845,9 @@ def apply_migrations():  # pragma: no cover
     default="ERROR",
 )
 @click.option("-v", "--verbose", count=True, help="Sets log level to DEBUG.")
+@click.option(
+    "-m", "--metrics", default=False, is_flag=True, help="Enables metrics logging."
+)
 @click.option(
     "-b",
     "--bot-token-file",
@@ -1763,6 +1902,7 @@ def apply_migrations():  # pragma: no cover
 def main(
     log_level,
     verbose,
+    metrics,
     bot_token_file,
     channel,
     auth_channels_file,
@@ -1784,7 +1924,8 @@ def main(
         token=get_token(bot_token_file),
         channels=auth_channels,
         db_url=database_url,
-        log_level=getattr(logging, "DEBUG" if verbose else log_level),
+        log_level="DEBUG" if verbose else log_level,
+        metrics=metrics,
     )
 
     if dev:
